/* Copyright (c) 2010 Michael Pippig
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pfft.h"
#include "ipfft.h"


INT PX(local_size_3d_guru)(
    const INT *n, const INT *ni, const INT *no,
    const PX(profile) *prf, const PX(permutation) *perm,
    INT howmany, const INT *block_col, const INT *block_row,
    const INT *num_gcells_below, const INT *num_gcells_above,
    MPI_Comm comm_cart_2d,
    INT *local_ni, INT *local_i_start,
    INT *local_no, INT *local_o_start,
    INT *alloc_local_data_in,
    PX(permutation)* perm_out
    )
{
  INT alloc_local_gcells;
  INT block_col_intern[3], block_row_intern[3];
  INT num_gcells_below_intern[3], num_gcells_above_intern[3];
  MPI_Comm comm_col, comm_row;
  PX(truncation) trc;

  PX(init_truncation)(ni, no, &trc);

  PX(split_cart_procmesh_2d)(comm_cart_2d, &comm_col, &comm_row);
  
  PX(evaluate_user_block_size_3d_truncated)(
      ni, no, perm, block_col, block_row, comm_col, comm_row,
      block_col_intern, block_row_intern);

  *alloc_local_data_in = PX(local_size_dft_3d_internal)(
      n, prf, &trc, perm, howmany, block_col_intern, block_row_intern, comm_col, comm_row,
      local_ni, local_i_start, local_no, local_o_start);

  MPI_Comm_free(&comm_col); MPI_Comm_free(&comm_row);

  /* switch to final data distribution */
  PX(copy_permutation)(perm, perm_out);
  PX(permute_fft_3d)(prf, perm_out);

  PX(evaluate_user_ghostcells)(
     num_gcells_below, num_gcells_above,
     num_gcells_below_intern, num_gcells_above_intern);

  alloc_local_gcells = PX(local_size_ghostcells_internal)(
      local_no, howmany, num_gcells_below_intern, num_gcells_above_intern);

  return MAX(alloc_local_gcells, *alloc_local_data_in);
}


PX(plan) PX(plan_guru_dft_3d)(
    const INT *n, const INT *ni, const INT *no,
    const PX(profile) *prf, const PX(permutation) *perm,
    INT howmany, const INT *block_col, const INT *block_row,
    const INT *num_gcells_below, const INT *num_gcells_above,
    C *data_in, C *data_out,
    MPI_Comm comm_cart_2d,
    int sign, unsigned fftw_flags
    )
{
  INT block_col_intern[3], block_row_intern[3];
  INT num_gcells_below_intern[3], num_gcells_above_intern[3];
  MPI_Comm comm_col, comm_row;
  PX(plan) fftplan=NULL;
  PX(truncation) trc;
  PX(permutation) perm_out;

  PX(init_truncation)(ni, no, &trc);

  PX(split_cart_procmesh_2d)(comm_cart_2d, &comm_col, &comm_row);
  PX(evaluate_user_block_size_3d_truncated)(
      ni, no, perm, block_col, block_row, comm_col, comm_row,
      block_col_intern, block_row_intern);

  fftplan = PX(plan_dft_3d_internal)(
      n, prf, &trc, perm, howmany, block_col_intern, block_row_intern,
      data_in, data_out, comm_col, comm_row, sign, fftw_flags);
  
  /* switch to final data distribution */
  PX(copy_permutation)(perm, &perm_out);
  PX(permute_fft_3d)(prf, &perm_out);
  
  PX(evaluate_user_ghostcells)(
     num_gcells_below, num_gcells_above,
     num_gcells_below_intern, num_gcells_above_intern);
  
  fftplan->gcplan = PX(plan_ghostcells_internal)(
      trc.no, &perm_out, 2*howmany, block_col_intern, block_row_intern,
      num_gcells_below_intern, num_gcells_above_intern,
      (R*) data_out, comm_col, comm_row, comm_cart_2d);
  
  MPI_Comm_free(&comm_col); MPI_Comm_free(&comm_row);
  return fftplan;
}


int PX(check_guru)(
    const INT *n, const INT *ni, const INT *no,
    const PX(profile) *prf, const PX(permutation) *perm,
    const INT *block_col, const INT *block_row, MPI_Comm comm_cart_2d,
    int verbose
    )
{
  /* check arbitrary profile
   * errorcode = 0 : everything is fine
   * errorcode = 1 : an error occurred
   * errorcode = 2 : warning, something could be wrong
   * set verbose=1 for detailed information
   */

  int retval = 0;
  INT block_col_intern[3], block_row_intern[3];
  MPI_Comm comm_col, comm_row;
  PX(truncation) trc;

  PX(init_truncation)(ni, no, &trc);

  retval = PX(check_cart_2d_internal)(comm_cart_2d, verbose);

  if(retval != 1){
    PX(split_cart_procmesh_2d)(comm_cart_2d, &comm_col, &comm_row);
    PX(evaluate_user_block_size_3d_truncated)(
        ni, no, perm, block_col, block_row, comm_col, comm_row,
        block_col_intern, block_row_intern);

    retval = PX(check_internal)(n, prf, &trc, perm,
        block_col_intern, block_row_intern, comm_col, comm_row, verbose);

    MPI_Comm_free(&comm_col); MPI_Comm_free(&comm_row);
  }

  return retval;
}


