/*
 * Copyright (c) 2010 Michael Pippig
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/* This file is based on ifftw3.h and ifftw3-mpi.h */


/* PFFT internal header file */
#ifndef __IPFFT_H__
#define __IPFFT_H__


/* Begin: This part is based on ifftw3.h */
#include <stdlib.h>		/* size_t */
#include <stdarg.h>		/* va_list */
#include <stddef.h>		/* ptrdiff_t */
#include <stdio.h>		/* fprintf */

#include <math.h>
#include <mpi.h>
#include <fftw3-mpi.h>

#define IPFFT_EXTERN extern

typedef ptrdiff_t INT;
/*
  integral type large enough to contain a stride (what ``int'' should
  have been in the first place.
*/

#define IF(x,a,b) ((x)?(a):(b))

#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))
#define ABS(x) (((x)>0)?(x):(-(x)))
#define SIGN(a) (((a)>=0)?1:-1)
#define MACRO_PLAIN_INDEX_3D_VECTOR(k, n)     ( k[2] + n[2]*(k[1] + n[1]*k[0]) )
#define MACRO_PLAIN_INDEX_3D(k0,k1,k2,n)      ( (k2) + n[2]*( (k1) + n[1]*(k0) ) )


/* all pfft identifiers start with pfft (or pfftf etc.) */
#define CONCAT(prefix, name) prefix ## name
#if defined(PFFT_SINGLE)
  typedef float R;
#  define PFFT_MPI_TYPE MPI_FLOAT
#  define PX(name) CONCAT(pfftf_, name)
#  define X(name)  CONCAT(fftwf_, name)
#elif defined(PFFT_LDOUBLE)
  typedef long double R;
#  define PFFT_MPI_TYPE MPI_LONG_DOUBLE
#  define PX(name) CONCAT(pfftl_, name)
#  define X(name)  CONCAT(fftwl_, name)
#else
  typedef double R;
#  define PFFT_MPI_TYPE MPI_DOUBLE
#  define PX(name) CONCAT(pfft_, name)
#  define X(name)  CONCAT(fftw_, name)
#endif

#define XM(name)  X(CONCAT(mpi_, name))
typedef X(complex) C;

#define INPLACE_ON_2ND_REMAP        (1U<< 0)
#define INPLACE_ON_1ST_REMAP        (1U<< 1)
#define STRIDES_CHANGE_ON_1ST_REMAP (1U<< 2)
#define STRIDES_CHANGE_ON_2ND_REMAP (1U<< 3)
#define FFT_ON_1ST_REMAP            (1U<< 4)
#define FFT_ON_2ND_REMAP            (1U<< 5)

#define PFFT_GC_BORDERS           (0U)
#define PFFT_GC_CORNERS           (1U<< 0)
#define PFFT_GC_TRAFO             (0U)
#define PFFT_GC_ADJOINT           (1U<< 1)
#define PFFT_GC_SENDRECV          (1U<< 0)
#define PFFT_GC_RMA               (1U<< 1)

#define FFT_DATA_INIT  ( 1000.0 / (i + j*ni[0] + k*ni[0]*ni[1] + 1) )
//#define FFT_DATA_INIT  (i + j*ni[0] + k*ni[0]*ni[1] + 1)
// #define FFT_DATA_INIT (10*(((double)rand())/RAND_MAX - 0.5))



#ifndef __PFFT_H__
typedef struct{
    INT ni[3];
    INT no[3];
    int is_transformed[3];
} PX(truncation);

typedef struct {
  int n[3];
  int np[3];
} PX(permutation);

typedef struct {
  unsigned fft_flag[3];
  int first_dims[3];
  unsigned remap_flag[3];
  unsigned pfft_flags;
} PX(profile);

typedef struct PX(plan_s) *PX(plan);
#endif /* ndef __PFFT_H__ */



typedef struct PX(serfft_trunc_plan_s) serfft_trunc_plan;
typedef struct PX(serfft_plan_s) serfft_plan;
typedef struct PX(trunc_plan_s) trunc_plan;
typedef struct PX(plan_s) pfft_plan_internal;
typedef struct PX(gcplan_s) gcplan_internal;
typedef struct PX(gcplan_s) *PX(gcplan);

struct PX(serfft_trunc_plan_s){
  INT local_ni_embed[3];
  INT local_n_embed[3];
  INT local_n_trunc[3];
  INT local_no_trunc[3];
  INT howmany;
  C *data_in;
  C *data_out;
  serfft_plan *serfft;
};

struct PX(serfft_plan_s){
  X(plan) plan[2];
};

struct PX(plan_s){
  X(plan) global_remap[3];
  serfft_trunc_plan *local_fft[3];
  PX(gcplan) gcplan;
};


typedef struct{
  R *realData;
  INT globalArraySize[3];
  INT localArraySize[3];
  INT blockSize[3];
  INT numGhostCellsBelow[3];
  INT numGhostCellsAbove[3];
  INT tupleSize;
} gcell_plan;


struct PX(gcplan_s){
  INT n[3];
  INT loc_n[3];
  INT gc_below[3];
  INT gc_above[3];
  INT ngc[3];
  INT ngc_left[3];
  INT ngc_right[3];
  INT ngc_lower[3];
  INT ngc_upper[3];
  INT ngc_back[3];
  INT ngc_front[3];
  int rank_left;
  int rank_right;
  int rank_lower;
  int rank_upper;
  int rank_back;
  int rank_front;
  INT blk[3];
  INT tuple;
  R *data;
  MPI_Comm comm_cart_1d[3];
  MPI_Group grp;
  MPI_Win win;
  unsigned alg_flag;
};


/* block.c */
/* TODO: search for other helpfull functions in FFTWs mpi/block.c */
INT PX(global_block_size)(
    INT arraySize, INT userBlockSize, int numProcs);
void PX(local_block_size_and_offset)(
    INT global_array_size, INT global_block_size, int which_block,
    INT *local_block_size, INT *local_block_start);
INT PX(local_block_size_shifted)(
    INT globalArraySize, INT userBlockSize, int shift, MPI_Comm commCart1d);
INT PX(local_block_offset)(
    INT arrayLength, INT globalBlockSize, int whichBlock);
INT PX(local_block_size)(
    INT arraySize, INT globalBlockSize, int whichBlock);
INT PX(num_blocks)(INT global_array_size, INT global_block_size);



/* utils.c */

void PX(permute)(
    int p0, int p1, int p2, int* q);
void PX(permute_vector)(
    unsigned fft_flag, int* permutation);
void PX(decompose_fft_trunc_3d)(
    const INT *n, const PX(truncation) *trc, const PX(permutation) *perm,
    const INT *block_col, const INT *block_row,
    MPI_Comm comm_col, MPI_Comm comm_row,
    INT *local_ni, INT *local_n, INT *local_no);
void PX(decompose_3d)(
    const INT *n, const PX(permutation) *perm,
    const INT *block_col, const INT *block_row, MPI_Comm comm_col, MPI_Comm comm_row,
    INT *local_n, INT *local_n_start);
void PX(decompose_1d)(
    INT arrayLength, INT blockSize, int whichBlock,
    INT* blockStart, INT* blockEnd);
int PX(pos_mod)(
    int dividend, int divisor);
int PX(get_mpi_cart_coord)(
    MPI_Comm commCart1d, int *coord);
void PX(evaluate_user_block_size_3d_truncated)(
    const INT *ni, const INT *no, const PX(permutation) *perm,
    const INT *block_col, const INT *block_row,
    MPI_Comm comm_col, MPI_Comm comm_row,
    INT *block_col_intern, INT *block_row_intern);
void PX(evaluate_user_ghostcells)(
    const INT *gcells_below, const INT *gcells_above,
    INT *gcells_below_intern, INT *gcells_above_intern);
void PX(copy_permutation)(
    const PX(permutation) *perm,
    PX(permutation) *copy);
int PX(remap_num_comm)(
    int first_dims, unsigned remap_flag, const PX(permutation) *perm);
void PX(split_cart_procmesh_2d)(
    MPI_Comm comm_cart_2d,
    MPI_Comm *comm_col, MPI_Comm* comm_row);
int PX(is_cart_procmesh_2d)(
    MPI_Comm comm_cart_2d);
int PX(create_procmesh_2d)(
    MPI_Comm comm, int np0, int np1,
    MPI_Comm *comm_cart_2d);
int PX(get_local_dim)(
    const PX(permutation) *perm);
void PX(get_permutation_output)(
    const PX(profile) *prf, const PX(permutation) *perm_in,
    PX(permutation) *perm_out);
void PX(message)(
    const char *string, MPI_Comm comm);
int PX(flag_active)(
    unsigned flags, unsigned search_flag);
int PX(flag_not_active)(
    unsigned flags, unsigned search_flag);
int PX(permute_inverse)(
    const int *perm, int search_item);
int PX(get_dim_to_comm_num)(
    const PX(permutation) *perm, int comm_num);
void PX(get_current_global_size)(
    const PX(truncation) *trc,
    INT *n_trunc);
int PX(is_illegal_plan)(
    pfft_plan fftplan);
void PX(init_truncation)(
    const INT *ni, const INT *no,
    PX(truncation) *trc);





void PX(pad_permuted_array_with_zeros_outside)(
    const INT *orig_array_size, const INT *zeros_below, const INT *zeros_above,
    const PX(permutation) *perm, INT tuple_size, R *real_data);
void PX(truncate_permuted_array_outside)(
    const INT *orig_array_size, const INT *trunc_below, const INT *trunc_above,
    const PX(permutation) *perm, INT tuple_size, R *real_data);
void PX(truncate_permuted_array_inside)(
    const INT *n, const INT *trunc_size, const INT *trunc_start,
    const PX(permutation) *perm, INT tuple_size,
    R *real_data);

void PX(pad_array_with_zeros_outside)(
    const INT *n, const INT *zeros_below, const INT *zeros_above,
    INT tuple_size, R *real_data);
void PX(pad_array_with_zeros_inside)(
    const INT *n, const INT *zeros_size, const INT *zeros_start,
    INT tuple_size, R *real_data);
void PX(truncate_array_outside)(
    const INT *n, const INT *trunc_below, const INT *trunc_above,
    INT tuple_size, R *real_data);
void PX(truncate_array_inside)(
    const INT *n, const INT *trunc_size, const INT *trunc_start,
    INT tuple_size, R *real_data);


/* malloc.c */
void PX(die)(
    const char *s, MPI_Comm comm);




/* transpose.c */

void PX(init_transpose_2d)(
    const INT* n, const INT* block,
    int first_dims, MPI_Comm comm_cart_1d, unsigned fftw_flags,
    INT* lo, INT* hi, int* permutation, INT* N, INT* blk);
X(plan) PX(plan_transpose_2d)(
    const INT *n, const INT *local_n,
    INT tupleSize, const INT *block, C *data_in, C *data_out,
    int first_dims, MPI_Comm comm_cart_1d, unsigned fftw_flags,
    PX(permutation) *perm);
void PX(permute_transpose_2d)(
    int first_dims, unsigned remap_flags,
    PX(permutation) *perm);
void PX(permute_transposed_in_or_out)(
    int first_dims,
    PX(permutation) *perm);
void PX(permute_std_transpose_2d)(
    int first_dims,
    PX(permutation) *perm);
INT PX(local_size_transpose_2d)(
    const INT *n, const INT *local_n, INT howmany, const INT *block,
    int first_dims, MPI_Comm comm_cart_1d, unsigned fftw_flags,
    PX(permutation) *perm);

/* dft-serial.c */

serfft_trunc_plan* PX(plan_dft_trunc_1d)(
    const INT *local_ni, const INT *local_n, const INT *local_no,
    INT howmany, C *data_in, C *data_out,
    int sign, unsigned transpose_flags, unsigned fftw_flags,
    PX(truncation) *trc, PX(permutation) *perm);
void PX(serfft_trunc_destroy_plan)(
    serfft_trunc_plan *plan);
void PX(serfft_trunc_execute)(
    serfft_trunc_plan *plan);
void PX(permute_fft_1d)(
    unsigned fft_flag, PX(permutation) *perm);
void PX(mark_performed_fft)(
    const PX(permutation) *perm,
    PX(truncation) *trc);
void PX(serfft_execute)(
    serfft_plan *serfft);

/* dft-parallel.c */

INT PX(local_size_fft_1d)(
    const INT *local_n, INT howmany, unsigned fft_flag,
    PX(truncation) *trc, PX(permutation) *perm);
INT PX(local_size_dft_3d_internal)(
    const INT *n, const PX(profile) *prf,
    const PX(truncation) *trc, const PX(permutation) *perm,
    INT tuple_size, const INT *block_col, const INT *block_row,
    MPI_Comm comm_col, MPI_Comm comm_row,
    INT *local_ni,  INT *local_i_start,
    INT *local_no, INT *local_o_start);
pfft_plan PX(plan_dft_3d_internal)(
    const INT *n, const PX(profile) *prf,
    const PX(truncation) *trc, const PX(permutation) *perm,
    INT tuple_size, const INT *block_col, const INT *block_row,
    C *data_in, C *data_out, MPI_Comm comm_col, MPI_Comm comm_row,
    int sign, unsigned fftw_flags);
int PX(check_internal)(
    const INT *n, const PX(profile) *prf,
    const PX(truncation) *trc, const PX(permutation) *perm,
    const INT *block_col, const INT *block_row, MPI_Comm comm_col, MPI_Comm comm_row,
    int verbose);
void PX(permute_fft_3d)(
    const PX(profile) *prf, PX(permutation) *perm);


/* gcells_plan.c */

INT PX(local_size_ghostcells_internal)(
    const INT *localArraySize, INT tupleSize,
    const INT *numGhostCellsBelow, const INT *numGhostCellsAbove);
PX(gcplan) PX(plan_ghostcells_internal)(
    const INT *n, const PX(permutation) *perm,
    INT tuple_size, const INT *block_col, const INT *block_row,
    const INT *gcells_below, const INT *gcells_above,
    R *data, MPI_Comm comm_col, MPI_Comm comm_row, MPI_Comm comm_cart_2d);
void PX(destroy_gcplan)(
    PX(gcplan) ths);
void PX(exchange_ghostcells_internal)(
    PX(gcplan) ths);
void PX(reduce_ghostcells_internal)(
    PX(gcplan) ths);


/* gcells_sendrecv.c */

void PX(exchange_ghostcells_internal_sendrecv)(
    PX(gcplan) ths);
void PX(reduce_ghostcells_internal_sendrecv)(
    PX(gcplan) ths);

/* gcells_RMA.c */

void PX(exchange_ghostcells_internal_RMA)(
    PX(gcplan) ths);
void PX(reduce_ghostcells_internal_RMA)(
    PX(gcplan) ths);




/* check.c */
int PX(check_cart_2d_internal)(
    MPI_Comm comm_cart_2d, int verbose);


#endif /* __IPFFT_H__ */
